# vim: set ts=8 sw=4 sts=4 et ai tw=79:
# sipzamine (SIP Examine)
# Matches SIP dialogs by time, regex or pcap packet filter; returns concise SIP
# packet info.
# Copyright (C) 2011-2015,2018 Walter Doekes, OSSO B.V.

__author__ = 'Walter Doekes'
__copyright__ = 'Copyright (C) Walter Doekes, OSSO B.V. 2011-2015,2018'
__licence__ = 'GPLv3+'
__version__ = '0.2_dev'
